{-
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
-}
module Xml where {

import qualified Show;
import qualified Nat;
import qualified List;
import qualified Util;
import qualified Term;
import qualified ProofChecker;
import qualified Polynomial;
import qualified Trs;
import qualified HOL;
import qualified Integer;
import qualified ErrorMonad;
import qualified Product_Type;

data Xml = Xml [Char] [([Char], [Char])] [Xml.Xml] (Maybe [Char]);

wspace :: [Char];
wspace = List.append Show.newline (List.append Show.tab " ");

trim :: [Char] -> [Char];
trim [] = [];
trim (c : s) = (if List.member c Xml.wspace then Xml.trim s else c : s);

eq_xml :: Xml.Xml -> Xml.Xml -> Bool;
eq_xml (Xml.Xml list1 list2 list3 option) (Xml.Xml list1' list2' list3' option')
  = list1 == list1' &&
      (list2 == list2' && (list3 == list3' && option == option'));

instance Eq Xml.Xml where {
  (==) = Xml.eq_xml;
};

data Xmldoc = Xmldoc [Char] Xml.Xml;

data XtcTrs =
  Xtctrs [(Term.Term [Char] [Char], Term.Term [Char] [Char])]
    [([Char], Nat.Nat)] (Maybe Xml.Xml);

dquote :: [Char];
dquote = "\"";

parseError :: forall a. [Char] -> [Char] -> Util.Eithera [Char] a;
parseError s allowed =
  ErrorMonad.errora
    (List.append "expected symbols "
      (List.append allowed
        (List.append " but found " (List.takea (Integer.nat 30) s))));

consume2 ::
  [[Char]] -> [[Char]] -> [Char] -> Util.Eithera [Char] ([Char], [Char]);
consume2 (t : asa) allowed s =
  (if List.mapa snd (List.zipa t s) == t
    then ErrorMonad.returna (t, Xml.trim (List.dropa (List.size_list t) s))
    else Xml.consume2 asa allowed s);
consume2 [] allowed s =
  Xml.parseError s
    (List.concata (List.mapa (\ sa -> List.append sa ",") allowed));

consume :: [[Char]] -> [Char] -> Util.Eithera [Char] ([Char], [Char]);
consume allowed s = Xml.consume2 allowed allowed s;

letters :: [Char];
letters = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_0123456789&;:";

showAtt :: ([Char], [Char]) -> [Char];
showAtt (a, v) =
  List.append a
    (List.append "=" (List.append Xml.dquote (List.append v Xml.dquote)));

digit2nat :: Char -> Util.Eithera [Char] Nat.Nat;
digit2nat x =
  (if x == '0' then ErrorMonad.returna Nat.Zero_nat
    else (if x == '1' then ErrorMonad.returna Nat.one_nat
           else (if x == '2' then ErrorMonad.returna (Integer.nat 2)
                  else (if x == '3' then ErrorMonad.returna (Integer.nat 3)
                         else (if x == '4'
                                then ErrorMonad.returna (Integer.nat 4)
                                else (if x == '5'
                                       then ErrorMonad.returna (Integer.nat 5)
                                       else (if x == '6'
      then ErrorMonad.returna (Integer.nat 6)
      else (if x == '7' then ErrorMonad.returna (Integer.nat 7)
             else (if x == '8' then ErrorMonad.returna (Integer.nat 8)
                    else (if x == '9' then ErrorMonad.returna (Integer.nat 9)
                           else ErrorMonad.errora
                                  (List.append [x] " is not a digit")))))))))));

string2natHelp :: Nat.Nat -> [Char] -> Util.Eithera [Char] Nat.Nat;
string2natHelp n (d : s) =
  ErrorMonad.bind (Xml.digit2nat d)
    (\ m ->
      Xml.string2natHelp (Nat.plus_nat (Nat.times_nat (Integer.nat 10) n) m) s);
string2natHelp n [] = ErrorMonad.returna n;

string2nat :: [Char] -> Util.Eithera [Char] Nat.Nat;
string2nat s =
  (if s == [] then ErrorMonad.errora "cannot convert empty string into number"
    else Xml.string2natHelp Nat.Zero_nat s);

showAtts :: [([Char], [Char])] -> [Char];
showAtts asa =
  List.foldra (\ a s -> List.append " " (List.append (Xml.showAtt a) s)) asa [];

showXMLindent :: [Char] -> Nat.Nat -> Xml.Xml -> [Char];
showXMLindent ind i (Xml.Xml n a c t) =
  List.concata
    [ind, "<", n, Xml.showAtts a,
      (if c == [] && t == Nothing then "/>"
        else List.concata
               [">", (if not (c == [])
                       then List.append
                              (List.concata
                                (List.mapa
                                  (Xml.showXMLindent
                                    (List.append ind (List.replicatea i ' ')) i)
                                  c))
                              ind
                       else []),
                 (if t == Nothing then [] else Nat.the t), "</", n, ">"])];

sshow_xml :: Xml.Xml -> [Char];
sshow_xml xml = Xml.trim (Xml.showXMLindent Show.newline (Integer.nat 2) xml);

xml2text :: [Char] -> Xml.Xml -> Util.Eithera [Char] [Char];
xml2text reqName (Xml.Xml name atts cs text) =
  (if name == reqName &&
        (atts == [] &&
          (cs == [] && (not (text == Nothing) && not (Nat.the text == []))))
    then ErrorMonad.returna (Nat.the text)
    else ErrorMonad.errora
           (List.concata
             ["could not extract text for ", reqName, " from  ", Show.newline,
               Xml.sshow_xml (Xml.Xml name atts cs text)]));

xml2nat :: [Char] -> Xml.Xml -> Util.Eithera [Char] Nat.Nat;
xml2nat reqName x = ErrorMonad.bind (Xml.xml2text reqName x) Xml.string2nat;

xml2var :: forall a. Xml.Xml -> Util.Eithera [Char] (Term.Term a [Char]);
xml2var x =
  ErrorMonad.bind (Xml.xml2text "var" x) (ErrorMonad.returna . Term.Var);

xml_sig :: [([Char], Nat.Nat)] -> Xml.Xml;
xml_sig fas =
  Xml.Xml "signature" []
    (List.mapa
      (\ (a @ (f, aa)) ->
        Xml.Xml "funcsym" []
          [Xml.Xml "name" [] [] (Just f),
            Xml.Xml "arity" [] [] (Just (Show.sshow_nat aa))]
          Nothing)
      fas)
    Nothing;

mapMonad ::
  forall a b. (Xml.Xml -> Util.Eithera a b) -> [Xml.Xml] -> Util.Eithera a [b];
mapMonad f (xml : rest) =
  (case f xml of {
    Util.Lefta a -> ErrorMonad.errora a;
    Util.Righta a ->
      (case Xml.mapMonad f rest of {
        Util.Lefta aa -> ErrorMonad.errora aa;
        Util.Righta asa -> ErrorMonad.returna (a : asa);
      });
  });
mapMonad uu [] = ErrorMonad.returna [];

splitterFun :: (Char -> Bool) -> [Char] -> ([Char], [Char]);
splitterFun uu [] = ([], []);
splitterFun allowed (c : s) =
  (if allowed c
    then let {
           a = Xml.splitterFun allowed s;
           (l, aa) = a;
         } in (c : l, aa)
    else ([], c : s));

parseText :: [Char] -> Util.Eithera [Char] (Maybe [Char], [Char]);
parseText s =
  let {
    a = Xml.splitterFun (\ c -> not ([c] == "<")) s;
    (l, r) = a;
  } in (if Xml.trim l == [] then ErrorMonad.returna (Nothing, r)
         else ErrorMonad.returna
                (Just (List.rev (Xml.trim (List.rev (Xml.trim l)))), r));

parseAttribute :: [Char] -> Util.Eithera [Char] ([Char], [Char]);
parseAttribute s =
  ErrorMonad.bind (Xml.consume [Xml.dquote] s)
    (\ (a @ (uu, s1)) ->
      let {
        aa = Xml.splitterFun (\ c -> not ([c] == Xml.dquote)) s1;
        (n, s2) = aa;
      } in ErrorMonad.bind (Xml.consume [Xml.dquote] s2)
             (\ (ab @ (uua, s3)) -> ErrorMonad.returna (n, s3)));

splitter :: [Char] -> [Char] -> ([Char], [Char]);
splitter allowed s = Xml.splitterFun (\ c -> List.member c allowed) s;

parseName :: [Char] -> Util.Eithera [Char] ([Char], [Char]);
parseName s =
  let {
    a = Xml.splitter Xml.letters s;
    (l, r) = a;
  } in (if l == [] then Xml.parseError s Xml.letters
         else ErrorMonad.returna (l, Xml.trim r));

parseAttributes :: [Char] -> Util.Eithera [Char] ([([Char], [Char])], [Char]);
parseAttributes (c : s) =
  (if List.member c "/>" then ErrorMonad.returna ([], c : s)
    else ErrorMonad.bind (Xml.parseName (c : s))
           (\ (a @ (n, s1)) ->
             ErrorMonad.bind (Xml.consume ["="] s1)
               (\ (aa @ (uu, s2)) ->
                 ErrorMonad.bind (Xml.parseAttribute s2)
                   (\ (ab @ (ac, s3)) ->
                     ErrorMonad.bind (Xml.parseAttributes s3)
                       (\ (b @ (nas, s4)) ->
                         ErrorMonad.returna ((n, ac) : nas, s4))))));
parseAttributes [] = ErrorMonad.returna ([], []);

parseXMLList :: [Char] -> Util.Eithera [Char] ([Xml.Xml], [Char]);
parseXMLList s =
  (if s == [] || List.takea (Integer.nat 2) s == "</"
    then ErrorMonad.returna ([], s)
    else ErrorMonad.bind (Xml.consume ["<"] s)
           (\ (a @ (uu, s1)) ->
             ErrorMonad.bind (Xml.parseName s1)
               (\ (aa @ (n, s2)) ->
                 ErrorMonad.bind (Xml.parseAttributes s2)
                   (\ (ab @ (nas, s3)) ->
                     ErrorMonad.bind (Xml.consume ["/>", ">"] s3)
                       (\ (ac @ (clo, s4)) ->
                         (if clo == "/>"
                           then ErrorMonad.bind (Xml.parseXMLList s4)
                                  (\ (ad @ (li, s5)) ->
                                    ErrorMonad.returna
                                      (Xml.Xml n nas [] Nothing : li, s5))
                           else ErrorMonad.bind (Xml.parseText s4)
                                  (\ (ad @ (t, s5)) ->
                                    ErrorMonad.bind (Xml.parseXMLList s5)
                                      (\ (ae @ (chi, s6)) ->
ErrorMonad.bind (Xml.consume [List.append "</" (List.append n ">")] s6)
  (\ (af @ (uua, s7)) ->
    ErrorMonad.bind (Xml.parseXMLList s7)
      (\ (ag @ (li, s8)) ->
        ErrorMonad.returna (Xml.Xml n nas chi t : li, s8)))))))))));

parseXML :: [Char] -> Util.Eithera [Char] (Xml.Xml, [Char]);
parseXML s =
  ErrorMonad.bind (Xml.consume ["<"] s)
    (\ (a @ (uu, s1)) ->
      ErrorMonad.bind (Xml.parseName s1)
        (\ (aa @ (n, s2)) ->
          ErrorMonad.bind (Xml.parseAttributes s2)
            (\ (ab @ (nas, s3)) ->
              ErrorMonad.bind (Xml.consume ["/>", ">"] s3)
                (\ (ac @ (clo, s4)) ->
                  (if clo == "/>"
                    then ErrorMonad.returna
                           (Xml.Xml n nas [] Nothing, Xml.trim s4)
                    else ErrorMonad.bind (Xml.parseText s4)
                           (\ (ad @ (t, s5)) ->
                             ErrorMonad.bind (Xml.parseXMLList s5)
                               (\ (ae @ (chi, s6)) ->
                                 ErrorMonad.bind
                                   (Xml.consume
                                     [List.append "</" (List.append n ">")] s6)
                                   (\ (af @ (uua, s7)) ->
                                     ErrorMonad.returna
                                       (Xml.Xml n nas chi t, s7)))))))));

scanUpto :: [Char] -> [Char] -> Util.Eithera [Char] ([Char], [Char]);
scanUpto end (c : s) =
  (if List.mapa snd (List.zipa end (c : s)) == end
    then ErrorMonad.returna
           (List.takea (List.size_list end) (c : s),
             Xml.trim (List.dropa (List.size_list end) (c : s)))
    else ErrorMonad.bind (Xml.scanUpto end s)
           (\ (a @ (res, rest)) -> ErrorMonad.returna (c : res, rest)));
scanUpto end [] =
  ErrorMonad.errora (List.append "did not find required token " end);

string2cint :: [Char] -> Util.Eithera [Char] Util.Cint;
string2cint s =
  (case s of {
    [] -> ErrorMonad.errora "cannot convert empty string into number";
    h : t ->
      let {
        a = (if h == '-' then (False, t) else (True, s));
        (pos, ss) = a;
      } in ErrorMonad.bind (Xml.string2nat ss)
             (\ n ->
               (if pos || Nat.eq_nat n Nat.Zero_nat
                 then ErrorMonad.returna (Util.Pos n)
                 else ErrorMonad.returna (Util.Neg n)));
  });

xml2cint :: [Char] -> Xml.Xml -> Util.Eithera [Char] Util.Cint;
xml2cint reqName x = ErrorMonad.bind (Xml.xml2text reqName x) Xml.string2cint;

xml2name :: Xml.Xml -> Util.Eithera [Char] [Char];
xml2name (Xml.Xml name atts cs text) =
  (if name == "name" &&
        (cs == [] && (not (text == Nothing) && not (Nat.the text == [])))
    then (if atts == [] || atts == [("sharp", "false")]
           then ErrorMonad.returna (Nat.the text)
           else ErrorMonad.errora
                  (List.append
                    (List.append "invalid attributes in " Show.newline)
                    (Xml.sshow_xml (Xml.Xml name atts cs text))))
    else ErrorMonad.errora
           (List.append (List.append "could not parse name from " Show.newline)
             (Xml.sshow_xml (Xml.Xml name atts cs text))));

xml1element ::
  forall a b.
    [Char] ->
      (Xml.Xml -> Util.Eithera [Char] a) ->
        (a -> b) -> Xml.Xml -> Util.Eithera [Char] b;
xml1element reqName p1 comb (Xml.Xml name atts cs text) =
  (if name == reqName &&
        (atts == [] &&
          (Nat.eq_nat (List.size_list cs) Nat.one_nat && text == Nothing))
    then ErrorMonad.bind (p1 (List.nth cs Nat.Zero_nat))
           (ErrorMonad.returna . comb)
    else ErrorMonad.errora
           (List.concata
             ["could not transform the following xml element (expected ",
               reqName, ")", Show.newline,
               Xml.sshow_xml (Xml.Xml name atts cs text)]));

xml1manyElements ::
  forall a b c.
    [Char] ->
      (Xml.Xml -> Util.Eithera [Char] a) ->
        (Xml.Xml -> Util.Eithera [Char] b) ->
          (a -> [b] -> c) -> Xml.Xml -> Util.Eithera [Char] c;
xml1manyElements reqName p1 p2 comb (Xml.Xml name atts cs text) =
  (if name == reqName && (atts == [] && (text == Nothing && not (cs == [])))
    then ErrorMonad.bind (p1 (List.hd cs))
           (\ a ->
             ErrorMonad.bind (Xml.mapMonad p2 (List.tl cs))
               (ErrorMonad.returna . comb a))
    else ErrorMonad.errora
           (List.concata
             ["could not transform the following xml element (expected ",
               reqName, ")", Show.newline,
               Xml.sshow_xml (Xml.Xml name atts cs text)]));

xmlOptionsHelp ::
  forall a.
    [(Xml.Xml -> Util.Eithera [Char] a)] ->
      [[Char]] -> Xml.Xml -> Util.Eithera [Char] a;
xmlOptionsHelp (p : ps) errs x =
  ErrorMonad.catcha (p x)
    (\ err -> Xml.xmlOptionsHelp ps (List.append errs [err]) x);
xmlOptionsHelp [] errs x =
  ErrorMonad.errora
    (List.append
      (List.concata ["no option found for ", Show.newline, Xml.sshow_xml x])
      (List.concata
        (List.mapa
          (\ err -> List.concata [Show.newline, "-------", Show.newline, err])
          errs)));

xmlOptions ::
  forall a.
    [(Xml.Xml -> Util.Eithera [Char] a)] -> Xml.Xml -> Util.Eithera [Char] a;
xmlOptions ps x = Xml.xmlOptionsHelp ps [] x;

xml2term :: Xml.Xml -> Util.Eithera [Char] (Term.Term [Char] [Char]);
xml2term x =
  Xml.xmlOptions
    [Xml.xml2var,
      Xml.xml1manyElements "funapp" Xml.xml2name
        (Xml.xml1element "arg" Xml.xml2term id) Term.Fun]
    x;

xml2elements ::
  forall a b c.
    [Char] ->
      (Xml.Xml -> Util.Eithera [Char] a) ->
        (Xml.Xml -> Util.Eithera [Char] b) ->
          (a -> b -> c) -> Xml.Xml -> Util.Eithera [Char] c;
xml2elements reqName p1 p2 comb (Xml.Xml name atts cs text) =
  (if name == reqName &&
        (atts == [] &&
          (Nat.eq_nat (List.size_list cs) (Integer.nat 2) && text == Nothing))
    then ErrorMonad.bind (p1 (List.nth cs Nat.Zero_nat))
           (\ a ->
             ErrorMonad.bind (p2 (List.nth cs Nat.one_nat))
               (ErrorMonad.returna . comb a))
    else ErrorMonad.errora
           (List.concata
             ["could not transform the following xml element (expected ",
               reqName, ")", Show.newline,
               Xml.sshow_xml (Xml.Xml name atts cs text)]));

xml2rule ::
  Xml.Xml ->
    Util.Eithera [Char] (Term.Term [Char] [Char], Term.Term [Char] [Char]);
xml2rule =
  Xml.xml2elements "rule" (Xml.xml1element "lhs" Xml.xml2term id)
    (Xml.xml1element "rhs" Xml.xml2term id) (\ a b -> (a, b));

xml_case ::
  forall t.
    ([Char] -> [([Char], [Char])] -> [Xml.Xml] -> Maybe [Char] -> t) ->
      Xml.Xml -> t;
xml_case f1 (Xml.Xml list1 list2 list3 option) = f1 list1 list2 list3 option;

xml_term ::
  forall c d. (Show.Sshow c, Show.Sshow d) => Term.Term c d -> Xml.Xml;
xml_term (Term.Fun f ts) =
  Xml.Xml "funapp" []
    (Xml.Xml "name" [] [] (Just (Show.sshow f)) :
      List.mapa (\ t -> Xml.Xml "arg" [] [Xml.xml_term t] Nothing) ts)
    Nothing;
xml_term (Term.Var x) = Xml.Xml "var" [] [] (Just (Show.sshow x));

fitLength :: forall a. a -> Nat.Nat -> [a] -> [a];
fitLength a (Nat.Suc n) (b : asa) = b : Xml.fitLength a n asa;
fitLength a (Nat.Suc v) [] = List.replicatea (Nat.Suc v) a;
fitLength uu Nat.Zero_nat uv = [];

xmlManyElements ::
  forall a b.
    [Char] ->
      (Xml.Xml -> Util.Eithera [Char] a) ->
        ([a] -> b) -> Xml.Xml -> Util.Eithera [Char] b;
xmlManyElements reqName p comb (Xml.Xml name atts cs text) =
  (if name == reqName && (atts == [] && text == Nothing)
    then ErrorMonad.bind (Xml.mapMonad p cs) (ErrorMonad.returna . comb)
    else ErrorMonad.errora
           (List.concata
             ["could not transform the following xml element (expected ",
               reqName, ")", Show.newline,
               Xml.sshow_xml (Xml.Xml name atts cs text)]));

xml2rules ::
  Xml.Xml ->
    Util.Eithera [Char] [(Term.Term [Char] [Char], Term.Term [Char] [Char])];
xml2rules = Xml.xmlManyElements "rules" Xml.xml2rule id;

xmlChange ::
  forall a b.
    (Xml.Xml -> Util.Eithera [Char] a) ->
      (a -> b) -> Xml.Xml -> Util.Eithera [Char] b;
xmlChange p f x = ErrorMonad.bind (p x) (\ a -> ErrorMonad.returna (f a));

xml_rules :: [(Term.Term [Char] [Char], Term.Term [Char] [Char])] -> Xml.Xml;
xml_rules lrs =
  Xml.Xml "rules" []
    (List.mapa
      (\ (a @ (l, r)) ->
        Xml.Xml "rule" []
          [Xml.Xml "lhs" [] [Xml.xml_term l] Nothing,
            Xml.Xml "rhs" [] [Xml.xml_term r] Nothing]
          Nothing)
      lrs)
    Nothing;

data XtcProblem =
  XTCProblem Xml.XtcTrs
    (Bool, ([Xml.Xml], ([([Char], [Char])], ([([Char], [Char])], [Char]))));

xml4elements ::
  forall a b c d e.
    [Char] ->
      (Xml.Xml -> Util.Eithera [Char] a) ->
        (Xml.Xml -> Util.Eithera [Char] b) ->
          (Xml.Xml -> Util.Eithera [Char] c) ->
            (Xml.Xml -> Util.Eithera [Char] d) ->
              (a -> b -> c -> d -> e) -> Xml.Xml -> Util.Eithera [Char] e;
xml4elements reqName p1 p2 p3 p4 comb (Xml.Xml name atts cs text) =
  (if name == reqName &&
        (atts == [] &&
          (Nat.eq_nat (List.size_list cs) (Integer.nat 4) && text == Nothing))
    then ErrorMonad.bind (p1 (List.nth cs Nat.Zero_nat))
           (\ a ->
             ErrorMonad.bind (p2 (List.nth cs Nat.one_nat))
               (\ b ->
                 ErrorMonad.bind (p3 (List.nth cs (Integer.nat 2)))
                   (\ c ->
                     ErrorMonad.bind (p4 (List.nth cs (Integer.nat 3)))
                       (ErrorMonad.returna . comb a b c))))
    else ErrorMonad.errora
           (List.concata
             ["could not transform the following xml element (expected ",
               reqName, ")", Show.newline,
               Xml.sshow_xml (Xml.Xml name atts cs text)]));

xml0elements :: forall a. [Char] -> a -> Xml.Xml -> Util.Eithera [Char] a;
xml0elements reqName comb (Xml.Xml name atts cs text) =
  (if name == reqName && (atts == [] && (cs == [] && text == Nothing))
    then ErrorMonad.returna comb
    else ErrorMonad.errora
           (List.concata
             ["could not transform the following xml element (expected ",
               reqName, ")", Show.newline,
               Xml.sshow_xml (Xml.Xml name atts cs text)]));

xml2context :: Xml.Xml -> Util.Eithera [Char] (Term.Ctxt [Char] [Char]);
xml2context x =
  Xml.xmlOptions
    [Xml.xml0elements "box" Term.Hole,
      Xml.xml4elements "funContext" Xml.xml2name
        (Xml.xmlManyElements "before" Xml.xml2term id) Xml.xml2context
        (Xml.xmlManyElements "after" Xml.xml2term id) Term.More]
    x;

xml2substitution ::
  Xml.Xml ->
    Util.Eithera [Char] [(Term.Term [Char] [Char], Term.Term [Char] [Char])];
xml2substitution =
  Xml.xmlManyElements "substitution"
    (Xml.xml2elements "substEntry" Xml.xml2var Xml.xml2term (\ a b -> (a, b)))
    id;

xml2manyElements ::
  forall a b c d.
    [Char] ->
      (Xml.Xml -> Util.Eithera [Char] a) ->
        (Xml.Xml -> Util.Eithera [Char] b) ->
          (Xml.Xml -> Util.Eithera [Char] c) ->
            (a -> b -> [c] -> d) -> Xml.Xml -> Util.Eithera [Char] d;
xml2manyElements reqName p1 p2 p3 comb (Xml.Xml name atts cs text) =
  (if name == reqName &&
        (atts == [] &&
          (text == Nothing &&
            Nat.less_eq_nat (Integer.nat 2) (List.size_list cs)))
    then ErrorMonad.bind (p1 (List.nth cs Nat.Zero_nat))
           (\ a ->
             ErrorMonad.bind (p2 (List.nth cs Nat.one_nat))
               (\ b ->
                 ErrorMonad.bind (Xml.mapMonad p3 (List.tl (List.tl cs)))
                   (ErrorMonad.returna . comb a b)))
    else ErrorMonad.errora
           (List.concata
             ["could not transform the following xml element (expected ",
               reqName, ")", Show.newline,
               Xml.sshow_xml (Xml.Xml name atts cs text)]));

xml2trsDisproof ::
  Xml.Xml -> Util.Eithera [Char] (ProofChecker.TRSDisproof [Char] [Char]);
xml2trsDisproof =
  Xml.xml2manyElements "loop" Xml.xml2substitution Xml.xml2context Xml.xml2term
    (\ sigma c ts -> ProofChecker.Loop (ts, (c, sigma)));

xml3elements ::
  forall a b c d.
    [Char] ->
      (Xml.Xml -> Util.Eithera [Char] a) ->
        (Xml.Xml -> Util.Eithera [Char] b) ->
          (Xml.Xml -> Util.Eithera [Char] c) ->
            (a -> b -> c -> d) -> Xml.Xml -> Util.Eithera [Char] d;
xml3elements reqName p1 p2 p3 comb (Xml.Xml name atts cs text) =
  (if name == reqName &&
        (atts == [] &&
          (Nat.eq_nat (List.size_list cs) (Integer.nat 3) && text == Nothing))
    then ErrorMonad.bind (p1 (List.nth cs Nat.Zero_nat))
           (\ a ->
             ErrorMonad.bind (p2 (List.nth cs Nat.one_nat))
               (\ b ->
                 ErrorMonad.bind (p3 (List.nth cs (Integer.nat 2)))
                   (ErrorMonad.returna . comb a b)))
    else ErrorMonad.errora
           (List.concata
             ["could not transform the following xml element (expected ",
               reqName, ")", Show.newline,
               Xml.sshow_xml (Xml.Xml name atts cs text)]));

xml2redpair ::
  forall a.
    (Xml.Xml -> Util.Eithera [Char] a) ->
      Xml.Xml -> Util.Eithera [Char] (Polynomial.RedPair a);
xml2redpair xml2sname =
  Xml.xmlOptions
    [Xml.xml1element "redPair"
       (Xml.xml2manyElements "interpretation"
         (Xml.xml1element "type" (Xml.xml0elements "linearPolynomial" ()) id)
         (Xml.xml1element "domain" (Xml.xml0elements "naturals" ()) id)
         (Xml.xml3elements "interpret" xml2sname (Xml.xml2nat "arity")
           (Xml.xml1element "function"
             (Xml.xml1manyElements "linearPolynomial"
               (Xml.xml1element "constant" (Xml.xml2cint "number") id)
               (Xml.xml1element "coefficient" (Xml.xml2nat "natural") id)
               (\ a b -> (a, b)))
             id)
           (\ f a poly ->
             (f, (fst poly, Xml.fitLength Nat.Zero_nat a (snd poly)))))
         (\ uu uua pia -> pia))
       Polynomial.NegPolo];

xml1to2elements ::
  forall a b c.
    [Char] ->
      (Xml.Xml -> Util.Eithera [Char] a) ->
        (Xml.Xml -> Util.Eithera [Char] b) ->
          (a -> Maybe b -> c) -> Xml.Xml -> Util.Eithera [Char] c;
xml1to2elements reqName p1 p2 comb (Xml.Xml name atts cs text) =
  (if name == reqName &&
        (atts == [] &&
          (Nat.less_eq_nat Nat.one_nat (List.size_list cs) &&
            (Nat.less_eq_nat (List.size_list cs) (Integer.nat 2) &&
              text == Nothing)))
    then ErrorMonad.bind (p1 (List.nth cs Nat.Zero_nat))
           (\ a ->
             (if Nat.eq_nat (List.size_list cs) (Integer.nat 2)
               then ErrorMonad.bind (p2 (List.nth cs Nat.one_nat))
                      (\ b -> ErrorMonad.returna (comb a (Just b)))
               else ErrorMonad.returna (comb a Nothing)))
    else ErrorMonad.errora
           (List.concata
             ["could not transform the following xml element (expected ",
               reqName, ")", Show.newline,
               Xml.sshow_xml (Xml.Xml name atts cs text)]));

xml2sharpName :: Xml.Xml -> Util.Eithera [Char] (Trs.Shp [Char]);
xml2sharpName (Xml.Xml name atts cs text) =
  (if name == "name" &&
        (cs == [] && (not (text == Nothing) && not (Nat.the text == [])))
    then (if atts == [] || atts == [("sharp", "false")]
           then ErrorMonad.returna (Trs.Plain (Nat.the text))
           else (if atts == [("sharp", "true")]
                  then ErrorMonad.returna (Trs.Sharp (Nat.the text))
                  else ErrorMonad.errora
                         (List.append
                           (List.append "invalid attributes in " Show.newline)
                           (Xml.sshow_xml (Xml.Xml name atts cs text)))))
    else ErrorMonad.errora
           (List.append
             (List.append "could not parse sharp name from " Show.newline)
             (Xml.sshow_xml (Xml.Xml name atts cs text))));

xml2sharpTerm ::
  Xml.Xml -> Util.Eithera [Char] (Term.Term (Trs.Shp [Char]) [Char]);
xml2sharpTerm x =
  Xml.xmlOptions
    [Xml.xml2var,
      Xml.xml1manyElements "funapp" Xml.xml2sharpName
        (Xml.xml1element "arg" Xml.xml2sharpTerm id) Term.Fun]
    x;

xml2sharpRule ::
  Xml.Xml ->
    Util.Eithera [Char]
      (Term.Term (Trs.Shp [Char]) [Char], Term.Term (Trs.Shp [Char]) [Char]);
xml2sharpRule =
  Xml.xml2elements "rule" (Xml.xml1element "lhs" Xml.xml2sharpTerm id)
    (Xml.xml1element "rhs" Xml.xml2sharpTerm id) (\ a b -> (a, b));

xml2sharpRules ::
  Xml.Xml ->
    Util.Eithera [Char]
      [(Term.Term (Trs.Shp [Char]) [Char], Term.Term (Trs.Shp [Char]) [Char])];
xml2sharpRules = Xml.xmlManyElements "rules" Xml.xml2sharpRule id;

xml2dpProof ::
  Xml.Xml -> Util.Eithera [Char] (ProofChecker.DPProof (Trs.Shp [Char]) [Char]);
xml2dpProof x =
  Xml.xmlOptions
    [Xml.xml0elements "pIsEmpty" ProofChecker.PisEmpty,
      Xml.xml3elements "redPairProc" (Xml.xml2redpair Xml.xml2sharpName)
        (Xml.xml1element "dps" Xml.xml2sharpRules id) Xml.xml2dpProof
        ProofChecker.RedPairProc,
      Xml.xml4elements "monoRedPairProc" (Xml.xml2redpair Xml.xml2sharpName)
        (Xml.xml1element "dps" Xml.xml2sharpRules id)
        (Xml.xml1element "trs" Xml.xml2sharpRules id) Xml.xml2dpProof
        ProofChecker.MonoRedPairProc,
      Xml.xmlManyElements "depGraphProc"
        (Xml.xml1to2elements "component"
          (Xml.xml1element "dps" Xml.xml2sharpRules id) Xml.xml2dpProof
          (\ dps prfOpt -> (prfOpt, dps)))
        ProofChecker.DepGraphProc]
    x;

xml2trsProof ::
  Xml.Xml -> Util.Eithera [Char] (ProofChecker.TRSProof [Char] [Char]);
xml2trsProof x =
  Xml.xmlOptions
    [Xml.xml0elements "rIsEmpty" ProofChecker.RisEmpty,
      Xml.xml2elements "dpTrans" (Xml.xml1element "dps" Xml.xml2sharpRules id)
        Xml.xml2dpProof ProofChecker.DPTrans,
      Xml.xml3elements "ruleRemoval" (Xml.xml2redpair Xml.xml2name)
        (Xml.xml1element "trs" Xml.xml2rules id) Xml.xml2trsProof
        ProofChecker.RuleRemoval]
    x;

xmldoc2proof ::
  Xml.Xmldoc ->
    Util.Eithera [Char]
      (Util.Eithera (ProofChecker.TRSProof [Char] [Char])
        (ProofChecker.TRSDisproof [Char] [Char]));
xmldoc2proof (Xml.Xmldoc header (Xml.Xml name atts cs text)) =
  (if name == "proof" &&
        (Nat.eq_nat (List.size_list cs) Nat.one_nat && text == Nothing)
    then Xml.xmlOptions
           [Xml.xmlChange Xml.xml2trsProof Util.Lefta,
             Xml.xmlChange Xml.xml2trsDisproof Util.Righta]
           (List.nth cs Nat.Zero_nat)
    else ErrorMonad.errora
           "invalid root element, expecting proof with exactly one child and no text");

stripXMLHeader :: [Char] -> Util.Eithera [Char] ([Char], [Char]);
stripXMLHeader s =
  (if List.takea (Integer.nat 2) (Xml.trim s) == "<?"
    then ErrorMonad.bind (Xml.scanUpto "?>" s)
           (\ (a @ (header1, s1)) ->
             ErrorMonad.bind (Xml.stripXMLHeader s1)
               (\ (aa @ (header2, s2)) ->
                 ErrorMonad.returna (List.append header1 header2, Xml.trim s2)))
    else ErrorMonad.returna ([], Xml.trim s));

parseXMLDoc :: [Char] -> Util.Eithera [Char] Xml.Xmldoc;
parseXMLDoc s =
  ErrorMonad.bind (Xml.stripXMLHeader s)
    (\ (a @ (header, rest)) ->
      ErrorMonad.bind (Xml.parseXML rest)
        (\ (aa @ (xml, end)) ->
          (if end == [] then ErrorMonad.returna (Xml.Xmldoc header xml)
            else ErrorMonad.errora
                   (List.append "there is text after root element: "
                     (List.append Show.newline end)))));

parseProof ::
  [Char] ->
    Util.Eithera [Char]
      (Util.Eithera (ProofChecker.TRSProof [Char] [Char])
        (ProofChecker.TRSDisproof [Char] [Char]));
parseProof s = ErrorMonad.bind (Xml.parseXMLDoc s) Xml.xmldoc2proof;

xmlGetName :: Xml.Xml -> [Char];
xmlGetName (Xml.Xml name uu uv uw) = name;

xml2funcsym :: Xml.Xml -> Util.Eithera [Char] ([Char], Nat.Nat);
xml2funcsym =
  Xml.xml2elements "funcsym" (Xml.xml2text "name") (Xml.xml2nat "arity")
    (\ a b -> (a, b));

xml2signature :: Xml.Xml -> Util.Eithera [Char] [([Char], Nat.Nat)];
xml2signature = Xml.xmlManyElements "signature" Xml.xml2funcsym id;

xml2to3elements ::
  forall a b c d.
    [Char] ->
      (Xml.Xml -> Util.Eithera [Char] a) ->
        (Xml.Xml -> Util.Eithera [Char] b) ->
          (Xml.Xml -> Util.Eithera [Char] c) ->
            (a -> b -> Maybe c -> d) -> Xml.Xml -> Util.Eithera [Char] d;
xml2to3elements reqName p1 p2 p3 comb (Xml.Xml name atts cs text) =
  (if name == reqName &&
        (atts == [] &&
          (Nat.less_eq_nat (Integer.nat 2) (List.size_list cs) &&
            (Nat.less_eq_nat (List.size_list cs) (Integer.nat 3) &&
              text == Nothing)))
    then ErrorMonad.bind (p1 (List.nth cs Nat.Zero_nat))
           (\ a ->
             ErrorMonad.bind (p2 (List.nth cs Nat.one_nat))
               (\ b ->
                 (if Nat.eq_nat (List.size_list cs) (Integer.nat 3)
                   then ErrorMonad.bind (p3 (List.nth cs (Integer.nat 2)))
                          (\ c -> ErrorMonad.returna (comb a b (Just c)))
                   else ErrorMonad.returna (comb a b Nothing))))
    else ErrorMonad.errora
           (List.concata
             ["could not transform the following xml element (expected ",
               reqName, ")", Show.newline,
               Xml.sshow_xml (Xml.Xml name atts cs text)]));

xml2xtcTrs :: Xml.Xml -> Util.Eithera [Char] Xml.XtcTrs;
xml2xtcTrs =
  Xml.xml2to3elements "trs" Xml.xml2rules Xml.xml2signature
    (\ x ->
      (if Xml.xmlGetName x == "comment" then ErrorMonad.returna x
        else ErrorMonad.errora
               (List.concata
                 ["unknown/unsupported third element in trs element",
                   Show.newline, Xml.sshow_xml x])))
    Xml.Xtctrs;

xml_xtcTrs :: Xml.XtcTrs -> Xml.Xml;
xml_xtcTrs (Xml.Xtctrs rls sig comment) =
  Xml.Xml "trs" []
    (List.append [Xml.xml_rules rls, Xml.xml_sig sig]
      (case comment of {
        Nothing -> [];
        Just c -> [c];
      }))
    Nothing;

xmlStartTerm :: Xml.Xml;
xmlStartTerm = Xml.Xml "startterm" [] [Xml.Xml "full" [] [] Nothing] Nothing;

xml_xtcProblem :: Xml.XtcProblem -> Xml.Xml;
xml_xtcProblem (Xml.XTCProblem xtcTrs (start, (ssm, (bef, (aft, header))))) =
  Xml.Xml "problem"
    (List.append bef (List.append [("type", "termination")] aft))
    (List.append [Xml.xml_xtcTrs xtcTrs, Xml.Xml "strategy" [] [] (Just "FULL")]
      (List.append (if start then [Xml.xmlStartTerm] else []) ssm))
    Nothing;

xtc2xmldoc :: Xml.XtcProblem -> Xml.Xmldoc;
xtc2xmldoc (Xml.XTCProblem xtcTrs (start, (ssm, (bef, (aft, header))))) =
  Xml.Xmldoc header
    (Xml.xml_xtcProblem
      (Xml.XTCProblem xtcTrs (start, (ssm, (bef, (aft, header))))));

xtcTrs_case ::
  forall t.
    ([(Term.Term [Char] [Char], Term.Term [Char] [Char])] ->
      [([Char], Nat.Nat)] -> Maybe Xml.Xml -> t)
      -> Xml.XtcTrs -> t;
xtcTrs_case f1 (Xml.Xtctrs list1 list2 option) = f1 list1 list2 option;

sshow_xmldoc :: Xml.Xmldoc -> [Char];
sshow_xmldoc (Xml.Xmldoc h x) = List.append h (Xml.sshow_xml x);

extractElement ::
  forall a. (Eq a) => a -> [a] -> Util.Eithera [Char] ([a], [a]);
extractElement a (b : bs) =
  (if HOL.eqop a b then ErrorMonad.returna ([], bs)
    else ErrorMonad.bind (Xml.extractElement a bs)
           (\ (aa @ (bef, aft)) -> ErrorMonad.returna (b : bef, aft)));
extractElement uu [] = ErrorMonad.errora "not found";

xmldoc2xtcProblem :: Xml.Xmldoc -> Util.Eithera [Char] Xml.XtcProblem;
xmldoc2xtcProblem (Xml.Xmldoc header xml) =
  let {
    (Xml.Xml name atts cs text) = xml;
  } in (if name == "problem" &&
             (Nat.less_eq_nat (Integer.nat 2) (List.size_list cs) &&
               (Nat.less_eq_nat Nat.one_nat (List.size_list atts) &&
                 text == Nothing))
         then (if not (Xml.eq_xml (List.nth cs Nat.one_nat)
                        (Xml.Xml "strategy" [] [] (Just "FULL")))
                then ErrorMonad.errora "unsupported or unknown strategy"
                else ErrorMonad.bind
                       (ErrorMonad.catcha
                         (Xml.extractElement ("type", "termination") atts)
                         (\ x ->
                           ErrorMonad.errora
                             "type termination attribute not found"))
                       (\ (a @ (bef, aft)) ->
                         ErrorMonad.bind
                           (Xml.xml2xtcTrs (List.nth cs Nat.Zero_nat))
                           (\ xtcTrs ->
                             let {
                               aa = (if Nat.less_nat (Integer.nat 2)
  (List.size_list cs)
  && Xml.eq_xml (List.nth cs (Integer.nat 2)) Xml.xmlStartTerm
                                      then (True, List.dropa (Integer.nat 3) cs)
                                      else (False,
     List.dropa (Integer.nat 2) cs));
                               (start, rest) = aa;
                             } in (if List.member "startterm"
(List.mapa Xml.xmlGetName rest)
                                    then ErrorMonad.errora
   "startterms unsupported"
                                    else ErrorMonad.returna
   (Xml.XTCProblem xtcTrs (start, (rest, (bef, (aft, header)))))))))
         else ErrorMonad.errora
                "root node must be problem and have at least the two initial children trs and strategy");

parseXTCProblem :: [Char] -> Util.Eithera [Char] Xml.XtcProblem;
parseXTCProblem s = ErrorMonad.bind (Xml.parseXMLDoc s) Xml.xmldoc2xtcProblem;

xtcProblem_case ::
  forall t.
    (Xml.XtcTrs ->
      (Bool, ([Xml.Xml], ([([Char], [Char])], ([([Char], [Char])], [Char])))) ->
        t)
      -> Xml.XtcProblem -> t;
xtcProblem_case f1 (Xml.XTCProblem xtcTrs x) = f1 xtcTrs x;

sshow_xtcProblem :: Xml.XtcProblem -> [Char];
sshow_xtcProblem p = Xml.sshow_xmldoc (Xml.xtc2xmldoc p);

}
